# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2026 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import os
import json
import logging
from pathlib import Path
from urllib.parse import urlparse

import click
from jsonschema import validate as jsonschema_validate
import yaml

from pygeoapi import __version__, l10n
from pygeoapi.models.openapi import OAPIFormat
from pygeoapi.util import to_json, yaml_load, remove_url_auth

LOGGER = logging.getLogger(__name__)

THISDIR = os.path.dirname(os.path.realpath(__file__))


def gen_asyncapi(cfg: dict) -> dict:
    """
    Generate an AsyncAPI document

    :param cfg: `dict` of pygeoapi configuration

    :returns: `dict` of AsyncAPI document
    """

    server_locales = l10n.get_locales(cfg)
    locale_ = server_locales[0]

    LOGGER.debug('Generating AsyncAPI document')

    title = l10n.translate(cfg['metadata']['identification']['title'], locale_)  # noqa
    description = l10n.translate(cfg['metadata']['identification']['description'], locale_)  # noqa
    tags = l10n.translate(cfg['metadata']['identification']['keywords'], locale_)  # noqa

    u = cfg['pubsub']['broker']['url']
    up = urlparse(u)
    protocol = up.scheme
    url = remove_url_auth(u).replace(f'{protocol}://', '')

    a = {
        'asyncapi': '3.0.0',
        'id': cfg['server']['url'],
        'defaultContentType': 'application/json',
        'info': {
            'version': __version__,
            'title': title,
            'description': description,
            'license': {
                'name': cfg['metadata']['license']['name'],
                'url': cfg['metadata']['license']['url']
            },
            'contact': {
                'name': cfg['metadata']['contact']['name'],
                'email': cfg['metadata']['contact']['email']
            },
            'tags': [{'name': tag} for tag in tags],
            'externalDocs': {
                'url': cfg['metadata']['identification']['url']
            },
        },
        'servers': {
            'default': {
                'host': url,
                'protocol': protocol,
                'description': description
            }
        },
        'channels': {},
        'operations': {}
    }
    if cfg['metadata']['contact']['url'].startswith('http'):
        a['info']['contact']['url'] = cfg['metadata']['contact']['url']

    if cfg['pubsub']['broker'].get('channel') is not None:
        channel_prefix = cfg['pubsub']['broker']['channel']
    else:
        channel_prefix = ''

    LOGGER.debug('Generating channels foreach collection')
    for key, value in cfg['resources'].items():
        if value['type'] not in ['collection']:
            LOGGER.debug('Skipping')
            continue

        title = l10n.translate(value['title'], locale_)
        channel_address = f'{channel_prefix}/collections/{key}'

        channel = {
            'description': title,
            'address': channel_address,
            'messages': {
                'DefaultMessage': {
                    'payload': {
                        '$ref': 'https://raw.githubusercontent.com/wmo-im/wis2-monitoring-events/refs/heads/main/schemas/cloudevents-v1.0.2.yaml'  # noqa
                    }
                }
            }
        }

        operation = {
            f'publish-{key}': {
                'action': 'send',
                'channel': {
                    '$ref': f'#/channels/notify-{key}'
                }
            },
            f'consume-{key}': {
                'action': 'receive',
                'channel': {
                    '$ref': f'#/channels/notify-{key}'
                }
            }
        }

        a['channels'][f'notify-{key}'] = channel
        a['operations'].update(operation)

    return a


def get_asyncapi(cfg, version='3.0'):
    """
    Stub to generate AsyncAPI Document

    :param cfg: configuration object
    :param version: version of AsyncAPI (default 3.0)

    :returns: AsyncAPI definition YAML dict
    """

    if version == '3.0':
        return gen_asyncapi(cfg)
    else:
        raise RuntimeError('AsyncAPI version not supported')


def validate_asyncapi_document(instance_dict):
    """
    Validate an AsyncAPI document against the AsyncAPI schema

    :param instance_dict: dict of AsyncAPI instance

    :returns: `bool` of validation
    """

    schema_file = os.path.join(
        THISDIR, 'resources', 'schemas', 'asyncapi', 'asyncapi-3.0.0.json')

    LOGGER.debug(f'Validating against {schema_file}')
    with open(schema_file) as fh2:
        schema_dict = json.load(fh2)
        jsonschema_validate(instance_dict, schema_dict)

        return True


def generate_asyncapi_document(cfg: dict, output_format: OAPIFormat):
    """
    Generate an AsyncAPI document from the configuration file

    :param cfg: `dict` of configuration
    :param output_format: output format for AsyncAPI document

    :returns: content of the AsyncAPI document in the output
              format requested
    """

    pretty_print = cfg['server'].get('pretty_print', False)

    if output_format == 'yaml':
        content = yaml.safe_dump(get_asyncapi(cfg), default_flow_style=False)
    else:
        content = to_json(get_asyncapi(cfg), pretty=pretty_print)
    return content


def load_asyncapi_document() -> dict:
    """
    Open AsyncAPI document from `PYGEOAPI_ASYNCAPI` environment variable

    :returns: `dict` of AsyncAPI document
    """

    pygeoapi_asyncapi = os.environ.get('PYGEOAPI_ASYNCAPI')

    if pygeoapi_asyncapi is None:
        LOGGER.debug('PYGEOAPI_ASYNCAPI environment not set')
        return {}

    if not os.path.exists(pygeoapi_asyncapi):
        msg = (f'AsyncAPI document {pygeoapi_asyncapi} does not exist.  '
               'Please generate before starting pygeoapi')
        LOGGER.warning(msg)
        return {}

    with open(pygeoapi_asyncapi, encoding='utf8') as ff:
        if pygeoapi_asyncapi.endswith(('.yaml', '.yml')):
            asyncapi_ = yaml_load(ff)
        else:  # JSON string, do not transform
            asyncapi_ = ff.read()

    return asyncapi_


@click.group()
def asyncapi():
    """AsyncAPI management"""
    pass


@click.command()
@click.pass_context
@click.argument('config_file', type=click.File(encoding='utf-8'))
@click.option('--format', '-f', 'format_', type=click.Choice(['json', 'yaml']),
              default='yaml', help='output format (json|yaml)')
@click.option('--output-file', '-of', type=click.File('w', encoding='utf-8'),
              help='Name of output file')
def generate(ctx, config_file, output_file, format_='yaml'):
    """Generate AsyncAPI Document"""

    if config_file is None:
        raise click.ClickException('--config/-c required')

    if isinstance(config_file, Path):
        with config_file.open(mode='r') as cf:
            cfg = yaml_load(cf)
    else:
        cfg = yaml_load(config_file)

    if 'pubsub' not in cfg:
        click.echo('pubsub not configured; aborting')
        ctx.exit(1)

    content = generate_asyncapi_document(cfg, format_)

    if output_file is None:
        click.echo(content)
    else:
        click.echo(f'Generating {output_file.name}')
        output_file.write(content)
        click.echo('Done')


@click.command()
@click.pass_context
@click.argument('asyncapi_file', type=click.File())
def validate(ctx, asyncapi_file):
    """Validate AsyncAPI Document"""

    if asyncapi_file is None:
        raise click.ClickException('--asyncapi/-o required')

    click.echo(f'Validating {asyncapi_file.name}')
    instance = yaml_load(asyncapi_file)
    validate_asyncapi_document(instance)
    click.echo('Valid AsyncAPI document')


asyncapi.add_command(generate)
asyncapi.add_command(validate)



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2023 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

"""Logging system"""

import logging
from logging.handlers import RotatingFileHandler
from logging.handlers import TimedRotatingFileHandler
import sys

LOGGER = logging.getLogger(__name__)


def setup_logger(logging_config: dict):
    """
    Setup configuration

    :param logging_config: logging specific configuration

    :returns: void (creates logging instance)
    """

    default_log_format = (
        '[%(asctime)s] {%(pathname)s:%(lineno)d} %(levelname)s - %(message)s'
    )
    default_date_format = '%Y-%m-%dT%H:%M:%SZ'

    log_format = logging_config.get('logformat', default_log_format)
    date_format = logging_config.get('dateformat', default_date_format)

    loglevels = {
        'CRITICAL': logging.CRITICAL,
        'ERROR': logging.ERROR,
        'WARNING': logging.WARNING,
        'INFO': logging.INFO,
        'DEBUG': logging.DEBUG,
        'NOTSET': logging.NOTSET
    }

    loglevel = loglevels[logging_config['level']]

    if 'logfile' in logging_config:
        rotation = logging_config.get('rotation')
        if rotation:
            rotate_mode = rotation.get('mode')

            rotate_backup_count = rotation.get('backup_count', 0)

            if rotate_mode == 'size':
                rotate_max_bytes = rotation.get('max_bytes', 0)

                logging.basicConfig(
                    handlers=[
                        RotatingFileHandler(
                            filename=logging_config['logfile'],
                            maxBytes=rotate_max_bytes,
                            backupCount=rotate_backup_count
                        )
                    ],
                    level=loglevel,
                    datefmt=date_format,
                    format=log_format,
                )
            elif rotate_mode == 'time':
                rotate_when = rotation.get('when', 'h')
                rotate_interval = rotation.get('interval', 1)

                logging.basicConfig(
                    handlers=[
                        TimedRotatingFileHandler(
                            filename=logging_config['logfile'],
                            when=rotate_when,
                            interval=rotate_interval,
                            backupCount=rotate_backup_count
                        )
                    ],
                    level=loglevel,
                    datefmt=date_format,
                    format=log_format
                )
            else:
                raise Exception(f'Invalid rotation mode:{rotate_mode}')
        else:
            logging.basicConfig(
                level=loglevel,
                datefmt=date_format,
                format=log_format,
                filename=logging_config['logfile']
            )
    else:
        logging.basicConfig(
            level=loglevel,
            datefmt=date_format,
            format=log_format,
            stream=sys.stdout
        )

    LOGGER.debug('Logging initialized')
    return



# =================================================================
#
# Authors: Bernhard Mallinger <bernhard.mallinger@eox.at>
#
# Copyright (c) 2024 Bernhard Mallinger
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================


from http import HTTPStatus


class GenericError(Exception):
    """Exception class where error codes and messages
    can be defined in custom error subclasses, so custom
    providers and processes can raise appropriate errors.
    """

    ogc_exception_code = 'NoApplicableCode'
    http_status_code = HTTPStatus.INTERNAL_SERVER_ERROR
    default_msg = 'Unknown error'

    def __init__(self, msg=None, *args, user_msg=None) -> None:
        # if only a user_msg is provided, use it as msg
        if user_msg and not msg:
            msg = user_msg
        super().__init__(msg, *args)
        self.user_msg = user_msg

    @property
    def message(self):
        return self.user_msg if self.user_msg else self.default_msg



# =================================================================
#
# Authors: Francesco Martinelli <francesco.martinelli@ingv.it>
#          Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2024 Francesco Martinelli
# Copyright (c) 2026 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

# Requires postgresql database structure.
# Create the database:
# e.g.
# CREATE DATABASE test
#   WITH TEMPLATE = template0
#   ENCODING = 'UTF8'
#   LOCALE = 'en_US.UTF-8';
# ALTER DATABASE test OWNER TO postgres;
#
# Import dump:
# psql -U postgres -h 127.0.0.1 -p 5432 test <
#   tests/data/postgres_manager_full_structure.backup.sql

import json
import logging
from pathlib import Path
from typing import Any, Tuple

from sqlalchemy import insert, update, delete
from sqlalchemy.orm import Session

from pygeoapi.process.base import (
    JobNotFoundError,
    JobResultNotFoundError,
    ProcessorGenericError
)
from pygeoapi.formats import FORMAT_TYPES, F_JSON, F_JSONLD
from pygeoapi.process.manager.base import BaseManager
from pygeoapi.provider.sql import (
    get_engine, get_table_model, store_db_parameters
)
from pygeoapi.util import JobStatus


LOGGER = logging.getLogger(__name__)


class PostgreSQLManager(BaseManager):
    """PostgreSQL Manager"""

    default_port = 5432

    def __init__(self, manager_def: dict):
        """
        Initialize object

        :param manager_def: manager definition

        :returns: `pygeoapi.process.manager.postgresql.PostgreSQLManager`
        """

        super().__init__(manager_def)
        self.is_async = True
        self.id_field = 'identifier'
        self.supports_subscribing = True
        self.connection = manager_def['connection']

        options = manager_def.get('options', {})
        store_db_parameters(self, manager_def['connection'], options)
        self._engine = get_engine(
            'postgresql+psycopg2',
            self.db_host,
            self.db_port,
            self.db_name,
            self.db_user,
            self._db_password,
            self.db_conn,
            **self.db_options
        )

        try:
            LOGGER.debug('Getting table model')
            self.table_model = get_table_model(
                'jobs',
                self.id_field,
                self.db_search_path,
                self._engine
            )
        except Exception as err:
            msg = 'Table model fetch failed'
            LOGGER.error(f'{msg}: {err}')
            raise ProcessorGenericError(msg)

    def get_jobs(self, status: JobStatus = None, limit=None, offset=None
                 ) -> dict:
        """
        Get jobs

        :param status: job status (accepted, running, successful,
                        failed, results) (default is all)
        :param limit: number of jobs to return
        :param offset: pagination offset

        :returns: dict of list of jobs (identifier, status, process identifier)
                  and numberMatched
        """

        LOGGER.debug('Querying for jobs')
        with Session(self._engine) as session:
            results = session.query(self.table_model)
            if status is not None:
                column = getattr(self.table_model, 'status')
                results = results.filter(column == status.value)

            jobs = [r.__dict__ for r in results.all()]
            return {
                'jobs': jobs,
                'numberMatched': len(jobs)
            }

    def add_job(self, job_metadata: dict) -> str:
        """
        Add a job

        :param job_metadata: `dict` of job metadata

        :returns: identifier of added job
        """

        LOGGER.debug('Adding job')
        with Session(self._engine) as session:
            try:
                session.execute(insert(self.table_model)
                                .values(**job_metadata))
                session.commit()
            except Exception as err:
                session.rollback()
                msg = 'Insert failed'
                LOGGER.error(f'{msg}: {err}')
                raise ProcessorGenericError(msg)

        return job_metadata['identifier']

    def update_job(self, job_id: str, update_dict: dict) -> bool:
        """
        Updates a job

        :param job_id: job identifier
        :param update_dict: `dict` of property updates

        :returns: `bool` of status result
        """

        rowcount = 0

        LOGGER.debug('Updating job')
        with Session(self._engine) as session:
            try:
                column = getattr(self.table_model, self.id_field)
                stmt = (
                    update(self.table_model)
                    .where(column == job_id)
                    .values(**update_dict)
                )
                result = session.execute(stmt)
                session.commit()
                rowcount = result.rowcount
            except Exception as err:
                session.rollback()
                msg = 'Update failed'
                LOGGER.error(f'{msg}: {err}')
                raise ProcessorGenericError(msg)

        return rowcount == 1

    def get_job(self, job_id: str) -> dict:
        """
        Get a single job

        :param job_id: job identifier

        :raises JobNotFoundError: if the job_id does not correspond to a
                                  known job
        :returns: `dict`  # `pygeoapi.process.manager.Job`
        """

        LOGGER.debug('Querying for job')
        with Session(self._engine) as session:
            results = session.query(self.table_model)
            column = getattr(self.table_model, self.id_field)
            results = session.query(self.table_model).filter(column == job_id)

            first = results.first()

            if first is not None:
                return first.__dict__
            else:
                raise JobNotFoundError()

    def delete_job(self, job_id: str) -> bool:
        """
        Deletes a job

        :param job_id: job identifier

        :raises JobNotFoundError: if the job_id does not correspond to a
                                  known job
        :return `bool` of status result
        """

        rowcount = 0

        # get result file if present for deletion
        job_result = self.get_job(job_id)
        location = job_result.get('location')

        LOGGER.debug('Deleting job')
        with Session(self._engine) as session:
            try:
                column = getattr(self.table_model, self.id_field)
                stmt = (
                    delete(self.table_model)
                    .where(column == job_id)
                )
                result = session.execute(stmt)
                session.commit()
                rowcount = result.rowcount
            except Exception as err:
                session.rollback()
                msg = 'Delete failed'
                LOGGER.error(f'{msg}: {err}')
                raise ProcessorGenericError(msg)

        # delete result file if present
        if None not in [location, self.output_dir]:
            try:
                Path(location).unlink()
            except FileNotFoundError:
                pass

        return rowcount == 1

    def get_job_result(self, job_id: str) -> Tuple[str, Any]:
        """
        Get a job's status, and actual output of executing the process

        :param job_id: job identifier

        :raises JobNotFoundError: if the job_id does not correspond to a
                                  known job
        :raises JobResultNotFoundError: if the job-related result cannot
                                        be returned
        :returns: `tuple` of mimetype and raw output
        """

        job_result = self.get_job(job_id)
        location = job_result.get('location')
        mimetype = job_result.get('mimetype')
        job_status = JobStatus[job_result['status']]

        if job_status != JobStatus.successful:
            # Job is incomplete
            return (None,)
        if not location:
            LOGGER.warning(f'job {job_id!r} -  unknown result location')
            raise JobResultNotFoundError()
        else:
            try:
                location = Path(location)
                if mimetype in (None, FORMAT_TYPES[F_JSON],
                                FORMAT_TYPES[F_JSONLD]):
                    with location.open('r', encoding='utf-8') as fh:
                        result = json.load(fh)
                else:
                    with location.open('rb') as fh:
                        result = fh.read()
            except (TypeError, FileNotFoundError, json.JSONDecodeError):
                raise JobResultNotFoundError()
            else:
                return mimetype, result

    def __repr__(self):
        return f'<PostgreSQLManager> {self.name}'



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#          Ricardo Garcia Silva <ricardo.garcia.silva@geobeyond.it>
#          Francesco Martinelli <francesco.martinelli@ingv.it>
#
# Copyright (c) 2024 Tom Kralidis
#           (c) 2023 Ricardo Garcia Silva
#           (c) 2026 Francesco Martinelli
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import collections
import json
import logging
from multiprocessing import dummy
from pathlib import Path
from typing import Any, Dict, Tuple, Optional, OrderedDict
import uuid

import requests

from pygeoapi.plugin import load_plugin
from pygeoapi.process.base import (
    BaseProcessor,
    JobNotFoundError,
    JobResultNotFoundError,
    ProcessorExecuteError,
    UnknownProcessError,
)
from pygeoapi.util import (
    get_current_datetime,
    is_request_allowed,
    JobStatus,
    ProcessExecutionMode,
    RequestedProcessExecutionMode,
    RequestedResponse,
    Subscriber
)

LOGGER = logging.getLogger(__name__)


class BaseManager:
    """generic Manager ABC"""
    processes: OrderedDict[str, Dict]

    def __init__(self, manager_def: dict):
        """
        Initialize object

        :param manager_def: manager definition

        :returns: `pygeoapi.process.manager.base.BaseManager`
        """

        self.name = manager_def['name']
        self.is_async = False
        self.supports_subscribing = False
        self.connection = manager_def.get('connection')
        self.output_dir = manager_def.get('output_dir')

        if self.output_dir is not None:
            self.output_dir = Path(self.output_dir)

        # Note: There are two different things named OrderedDict here - one
        # is coming from typing.OrderedDict (type annotation), the other is
        # coming from collections.OrderedDict (actual type we want to use here)
        # - this will not be needed anymore when pygeoapi moves to requiring
        # Python 3.9 as the minimum supported Python version
        self.processes = collections.OrderedDict()
        for id_, process_conf in manager_def.get('processes', {}).items():
            self.processes[id_] = dict(process_conf)

    def get_processor(self, process_id: str) -> BaseProcessor:
        """Instantiate a processor.

        :param process_id: Identifier of the process

        :raises UnknownProcessError: if the processor cannot be created
        :returns: instance of the processor
        """

        try:
            process_conf = self.processes[process_id]
        except KeyError as err:
            raise UnknownProcessError('Invalid process identifier') from err
        else:
            pp = load_plugin('process', process_conf['processor'])
            pp.allow_internal_requests = process_conf.get(
                'allow_internal_requests', False)

            return pp

    def get_jobs(self,
                 status: JobStatus = None,
                 limit: Optional[int] = None,
                 offset: Optional[int] = None
                 ) -> dict:
        """
        Get process jobs, optionally filtered by status

        :param status: job status (accepted, running, successful,
                       failed, results) (default is all)
        :param limit: number of jobs to return
        :param offset: pagination offset

        :returns: dict of list of jobs (identifier, status, process identifier)
                  and numberMatched
        """

        raise NotImplementedError()

    def add_job(self, job_metadata: dict) -> str:
        """
        Add a job

        :param job_metadata: `dict` of job metadata

        :returns: `str` added job identifier
        """

        raise NotImplementedError()

    def update_job(self, job_id: str, update_dict: dict) -> bool:
        """
        Updates a job

        :param job_id: job identifier
        :param update_dict: `dict` of property updates

        :returns: `bool` of status result
        """

        raise NotImplementedError()

    def get_job(self, job_id: str) -> dict:
        """
        Get a job (!)

        :param job_id: job identifier

        :raises JobNotFoundError: if the job_id does not correspond to a
                                  known job
        :returns: `dict` of job result
        """

        raise JobNotFoundError()

    def get_job_result(self, job_id: str) -> Tuple[str, Any]:
        """
        Returns the actual output from a completed process

        :param job_id: job identifier

        :raises JobNotFoundError: if the job_id does not correspond to a
                                  known job
        :raises JobResultNotFoundError: if the job-related result cannot
                                         be returned
        :returns: `tuple` of mimetype and raw output
        """

        raise JobResultNotFoundError()

    def delete_job(self, job_id: str) -> bool:
        """
        Deletes a job and associated results/outputs

        :param job_id: job identifier

        :raises JobNotFoundError: if the job_id does not correspond to a
                                   known job
        :returns: `bool` of status result
        """

        raise JobNotFoundError()

    def _execute_handler_async(self, p: BaseProcessor, job_id: str,
                               data_dict: dict,
                               requested_outputs: Optional[dict] = None,
                               subscriber: Optional[Subscriber] = None,
                               requested_response: Optional[RequestedResponse] = RequestedResponse.raw.value  # noqa
                               ) -> Tuple[str, None, JobStatus]:
        """
        This private execution handler executes a process in a background
        thread using `multiprocessing.dummy`

        https://docs.python.org/3/library/multiprocessing.html#module-multiprocessing.dummy  # noqa

        :param p: `pygeoapi.process` object
        :param job_id: job identifier
        :param data_dict: `dict` of data parameters
        :param requested_outputs: `dict` optionally specifying the subset of
                                  required outputs - defaults to all outputs.
                                  The value of any key may be an object and
                                  include the property `transmissionMode`
                                  (defaults to `value`)
                                  Note: 'optional' is for backward
                                  compatibility.
        :param subscriber: optional `Subscriber` specifying callback URLs
        :param requested_response: `RequestedResponse` optionally specifying
                                   raw or document (default is `raw`)

        :returns: tuple of None (i.e. initial response payload)
                  and JobStatus.accepted (i.e. initial job status)
        """

        args = (p, job_id, data_dict, requested_outputs, subscriber,
                requested_response)

        _process = dummy.Process(target=self._execute_handler_sync, args=args)
        _process.start()

        return 'application/json', None, JobStatus.accepted

    def _execute_handler_sync(self, p: BaseProcessor, job_id: str,
                              data_dict: dict,
                              requested_outputs: Optional[dict] = None,
                              subscriber: Optional[Subscriber] = None,
                              requested_response: Optional[RequestedResponse] = RequestedResponse.raw.value  # noqa
                              ) -> Tuple[str, Any, JobStatus]:
        """
        Synchronous execution handler

        If the manager has defined `output_dir`, then the result
        will be written to disk
        output store. There is no clean-up of old process outputs.

        :param p: `pygeoapi.process` object
        :param job_id: job identifier
        :param data_dict: `dict` of data parameters
        :param requested_outputs: `dict` optionally specifying the subset of
                                  required outputs - defaults to all outputs.
                                  The value of any key may be an object and
                                  include the property `transmissionMode`
                                  (defaults to `value`)
                                  Note: 'optional' is for backward
                                  compatibility.
        :param subscriber: optional `Subscriber` specifying callback URLs
        :param requested_response: `RequestedResponse` optionally specifying
                                   raw or document (default is `raw`)

        :returns: tuple of MIME type, response payload and status
        """

        extra_execute_parameters = {}

        # only pass requested_outputs if supported,
        # otherwise this breaks existing processes
        if p.supports_outputs:
            extra_execute_parameters['outputs'] = requested_outputs

        self._send_in_progress_notification(subscriber)

        try:
            if self.output_dir is not None:
                filename = f"{p.metadata['id']}-{job_id}"
                job_filename = self.output_dir / filename
            else:
                job_filename = None

            current_status = JobStatus.running
            jfmt, outputs = p.execute(data_dict, **extra_execute_parameters)

            if isinstance(outputs, bytes):
                outputs = outputs.decode('utf-8')

            if requested_response == RequestedResponse.document.value:
                outputs = {
                    'outputs': [outputs]
                }

            self.update_job(job_id, {
                'updated': get_current_datetime(),
                'status': current_status.value,
                'message': 'Writing job output',
                'progress': 95
            })

            if self.output_dir is not None:
                LOGGER.debug(f'writing output to {job_filename}')
                if isinstance(outputs, (dict, list)):
                    mode = 'w'
                    data = json.dumps(outputs, sort_keys=True, indent=4)
                    encoding = 'utf-8'
                elif isinstance(outputs, bytes):
                    mode = 'wb'
                    data = outputs
                    encoding = None
                elif isinstance(outputs, str):
                    mode = 'w'
                    data = outputs
                    encoding = None
                with job_filename.open(mode=mode, encoding=encoding) as fh:
                    fh.write(data)

            current_status = JobStatus.successful

            job_update_metadata = {
                'finished': get_current_datetime(),
                'updated': get_current_datetime(),
                'status': current_status.value,
                'location': str(job_filename),
                'mimetype': jfmt,
                'message': 'Job complete',
                'progress': 100
            }

            self.update_job(job_id, job_update_metadata)
            self._send_success_notification(subscriber, outputs=outputs)

        except Exception as err:
            # TODO assess correct exception type and description to help users
            # NOTE, the /results endpoint should return the error HTTP status
            # for jobs that failed, the specification says that failing jobs
            # must still be able to be retrieved with their error message
            # intact, and the correct HTTP error status at the /results
            # endpoint, even if the /result endpoint correctly returns the
            # failure information (i.e. what one might assume is a 200
            # response).

            current_status = JobStatus.failed
            code = 'InvalidParameterValue'
            outputs = {
                'type': code,
                'code': code,
                'description': f'Error executing process: {err}'
            }
            LOGGER.exception(err)
            job_metadata = {
                'finished': get_current_datetime(),
                'updated': get_current_datetime(),
                'status': current_status.value,
                'location': None,
                'mimetype': 'application/octet-stream',
                'message': f'{code}: {outputs["description"]}'
            }

            jfmt = 'application/json'

            self.update_job(job_id, job_metadata)
            self._send_failed_notification(subscriber)

        return jfmt, outputs, current_status

    def execute_process(
            self,
            process_id: str,
            data_dict: dict,
            execution_mode: Optional[RequestedProcessExecutionMode] = None,
            requested_outputs: Optional[dict] = None,
            subscriber: Optional[Subscriber] = None,
            requested_response: Optional[RequestedResponse] = RequestedResponse.raw.value  # noqa
    ) -> Tuple[str, Any, JobStatus, Optional[Dict[str, str]]]:
        """
        Default process execution handler

        :param process_id: process identifier
        :param data_dict: `dict` of data parameters
        :param execution_mode: `str` optionally specifying sync or async
                               processing.
        :param requested_outputs: `dict` optionally specifying the subset of
                                  required outputs - defaults to all outputs.
                                  The value of any key may be an object and
                                  include the property `transmissionMode`
                                  (default is `value`)
                                  Note: 'optional' is for backward
                                  compatibility.
        :param subscriber: `Subscriber` optionally specifying callback urls
        :param requested_response: `RequestedResponse` optionally specifying
                                   raw or document (default is `raw`)


        :raises UnknownProcessError: if the input process_id does not
                                     correspond to a known process
        :returns: tuple of job_id, MIME type, response payload, status and
                  optionally additional HTTP headers to include in the final
                  response
        """

        job_id = str(uuid.uuid1())
        self.processor = self.get_processor(process_id)
        self.processor.set_job_id(job_id)
        extra_execute_handler_parameters = {
            'requested_response': requested_response
        }

        job_control_options = self.processor.metadata.get(
            'jobControlOptions', [])

        if execution_mode == RequestedProcessExecutionMode.respond_async:
            # client wants async - do we support it?
            process_supports_async = (
                ProcessExecutionMode.async_execute.value in job_control_options
                )
            if self.is_async and process_supports_async:
                LOGGER.debug('Asynchronous execution')
                handler = self._execute_handler_async
                response_headers = {
                    'Preference-Applied': (
                        RequestedProcessExecutionMode.respond_async.value)
                }
            else:
                LOGGER.debug('Synchronous execution')
                handler = self._execute_handler_sync
                response_headers = {
                    'Preference-Applied': (
                        RequestedProcessExecutionMode.wait.value)
                }
        else:  # client has no preference or clients wants sync
            # do we support sync?
            process_supports_sync = (
                ProcessExecutionMode.sync_execute.value in job_control_options
                )
            if not process_supports_sync:
                LOGGER.debug('Asynchronous execution')
                handler = self._execute_handler_async
                response_headers = {
                    'Preference-Applied': (
                        RequestedProcessExecutionMode.respond_async.value)
                }
            else:
                # according to OAPI - Processes spec we ought to
                # respond with sync
                LOGGER.debug('Synchronous execution')
                handler = self._execute_handler_sync
                if execution_mode == RequestedProcessExecutionMode.wait:
                    response_headers = None
                else:
                    response_headers = {
                        'Preference-Applied': (
                            RequestedProcessExecutionMode.wait.value)
                    }

        # Add Job before returning any response.
        current_status = JobStatus.accepted
        job_metadata = {
            'type': 'process',
            'identifier': job_id,
            'process_id': process_id,
            'created': get_current_datetime(),
            'started': get_current_datetime(),
            'updated': get_current_datetime(),
            'finished': None,
            'status': current_status.value,
            'location': None,
            'mimetype': 'application/octet-stream',
            'message': 'Job accepted and ready for execution',
            'progress': 5
        }
        self.add_job(job_metadata)

        # only pass subscriber if supported, otherwise this breaks
        # existing managers
        if self.supports_subscribing:
            extra_execute_handler_parameters['subscriber'] = subscriber

        # TODO: handler's response could also be allowed to include more HTTP
        # headers
        mime_type, outputs, status = handler(
            self.processor,
            job_id,
            data_dict,
            requested_outputs,
            **extra_execute_handler_parameters)

        return job_id, mime_type, outputs, status, response_headers

    def _send_in_progress_notification(self, subscriber: Optional[Subscriber]):
        if subscriber and subscriber.in_progress_uri:
            self.__do_subscriber_request(subscriber.in_progress_uri)

    def _send_success_notification(
            self, subscriber: Optional[Subscriber], outputs: Any
    ):
        if subscriber and subscriber.success_uri:
            self.__do_subscriber_request(subscriber.success_uri, outputs)

    def _send_failed_notification(self, subscriber: Optional[Subscriber]):
        if subscriber and subscriber.failed_uri:
            self.__do_subscriber_request(subscriber.failed_uri)

    def __do_subscriber_request(self, url: str, data: dict = {}) -> None:
        """
        Helper function to execute a subscriber URL via HTTP POST

        :param url: `str` of URL
        :param data: `dict` of request payload

        :returns: `None`
        """

        if not is_request_allowed(url, self.processor.allow_internal_requests):
            msg = 'URL not allowed'
            LOGGER.error(f'{msg}: {url}')
            raise ProcessorExecuteError(msg)

        response = requests.post(url, json=data)
        LOGGER.debug(
            f'Response: {response.status_code}'
        )

    def __repr__(self):
        return f'<BaseManager> {self.name}'


def get_manager(config: Dict) -> BaseManager:
    """Instantiate process manager from the supplied configuration.

    :param config: pygeoapi configuration

    :returns: The pygeoapi process manager object
    """
    manager_conf = config.get('server', {}).get(
        'manager',
        {
            'name': 'Dummy',
            'connection': None,
            'output_dir': None
        }
    )
    processes_conf = {}
    for id_, resource_conf in config.get('resources', {}).items():
        if resource_conf.get('type') == 'process':
            processes_conf[id_] = resource_conf
    manager_conf['processes'] = processes_conf
    if manager_conf.get('name') == 'Dummy':
        LOGGER.info('Starting dummy manager')
    return load_plugin('process_manager', manager_conf)



# =================================================================

# Authors: Tom Kralidis <tomkralidis@gmail.com>
#          Francesco Bartoli <xbartolone@gmail.com>
#          Sander Schaminee <sander.schaminee@geocat.net>
#          John A Stevenson <jostev@bgs.ac.uk>
#          Colin Blackburn <colb@bgs.ac.uk>
#          Ricardo Garcia Silva <ricardo.garcia.silva@geobeyond.it>
#          Bernhard Mallinger <bernhard.mallinger@eox.at>
#          Francesco Martinelli <francesco.martinelli@ingv.it>
#
# Copyright (c) 2026 Tom Kralidis
# Copyright (c) 2025 Francesco Bartoli
# Copyright (c) 2022 John A Stevenson and Colin Blackburn
# Copyright (c) 2023 Ricardo Garcia Silva
# Copyright (c) 2024 Bernhard Mallinger
# Copyright (c) 2026 Francesco Martinelli
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================


import json
import logging
import urllib.parse
from copy import deepcopy
from datetime import datetime, timezone
from http import HTTPStatus
from typing import Tuple

from pygeoapi import l10n
from pygeoapi.formats import FORMAT_TYPES, F_HTML, F_JSON, F_JSONLD
from pygeoapi.api import evaluate_limit
from pygeoapi.api.pubsub import publish_message
from pygeoapi.process.base import (
    JobNotFoundError,
    JobResultNotFoundError,
    ProcessorExecuteError,
)
from pygeoapi.process.manager.base import get_manager, Subscriber
from pygeoapi.util import (
    json_serial, render_j2_template, JobStatus, RequestedProcessExecutionMode,
    to_json, DATETIME_FORMAT)

from . import APIRequest, API, SYSTEM_LOCALE

LOGGER = logging.getLogger(__name__)

CONFORMANCE_CLASSES = [
    'http://www.opengis.net/spec/ogcapi-processes-1/1.0/conf/ogc-process-description',  # noqa
    'http://www.opengis.net/spec/ogcapi-processes-1/1.0/conf/core',
    'http://www.opengis.net/spec/ogcapi-processes-1/1.0/conf/json',
    'http://www.opengis.net/spec/ogcapi-processes-1/1.0/conf/oas30',
    'http://www.opengis.net/spec/ogcapi-processes-1/1.0/conf/callback'
]


def describe_processes(api: API, request: APIRequest,
                       process: str | None = None
                       ) -> Tuple[dict, int, str]:
    """
    Provide processes metadata

    :param request: A request object
    :param process: process identifier, defaults to None to obtain
                    information about all processes

    :returns: tuple of headers, status code, content
    """

    processes = []

    headers = request.get_response_headers(**api.api_headers)

    if process is not None:
        if process not in api.manager.processes.keys():
            msg = 'Identifier not found'
            return api.get_exception(
                HTTPStatus.NOT_FOUND, headers,
                request.format, 'NoSuchProcess', msg)

    if len(api.manager.processes) > 0:
        if process is not None:
            relevant_processes = [process]
        else:
            LOGGER.debug('Processing limit parameter')
            if api.config['server'].get('limit') is not None:
                msg = ('server.limit is no longer supported! '
                       'Please use limits at the server or collection '
                       'level (RFC5)')
                LOGGER.warning(msg)
            try:
                limit = evaluate_limit(request.params.get('limit'),
                                       api.config['server'].get('limits', {}),
                                       {})
                relevant_processes = list(api.manager.processes)[:limit]
            except ValueError as err:
                return api.get_exception(
                    HTTPStatus.BAD_REQUEST, headers, request.format,
                    'InvalidParameterValue', str(err))

        for key in relevant_processes:
            p = api.manager.get_processor(key)
            p2 = l10n.translate_struct(deepcopy(p.metadata),
                                       request.locale)
            p2['id'] = key

            if process is None:
                p2.pop('inputs')
                p2.pop('outputs')
                p2.pop('example', None)

            p2['jobControlOptions'] = ['sync-execute']
            if api.manager.is_async:
                p2['jobControlOptions'].append('async-execute')

            p2['outputTransmission'] = ['value']

            p2['links'] = p2.get('links', [])

            jobs_url = f"{api.base_url}/jobs"
            process_url = f"{api.base_url}/processes/{key}"

            # TODO translation support
            link = {
                'type': FORMAT_TYPES[F_JSON],
                'rel': request.get_linkrel(F_JSON),
                'href': f'{process_url}?f={F_JSON}',
                'title': l10n.translate('Process description as JSON', request.locale),  # noqa
                'hreflang': api.default_locale
            }
            p2['links'].append(link)

            link = {
                'type': FORMAT_TYPES[F_HTML],
                'rel': request.get_linkrel(F_HTML),
                'href': f'{process_url}?f={F_HTML}',
                'title': l10n.translate('Process description as HTML', request.locale),  # noqa
                'hreflang': api.default_locale
            }
            p2['links'].append(link)

            link = {
                'type': FORMAT_TYPES[F_HTML],
                'rel': 'http://www.opengis.net/def/rel/ogc/1.0/job-list',
                'href': f'{jobs_url}?f={F_HTML}',
                'title': l10n.translate('Jobs list as HTML', request.locale),  # noqa
                'hreflang': api.default_locale
            }
            p2['links'].append(link)

            link = {
                'type': FORMAT_TYPES[F_JSON],
                'rel': 'http://www.opengis.net/def/rel/ogc/1.0/job-list',
                'href': f'{jobs_url}?f={F_JSON}',
                'title': l10n.translate('Jobs list as JSON', request.locale),  # noqa
                'hreflang': api.default_locale
            }
            p2['links'].append(link)

            link = {
                'type': FORMAT_TYPES[F_JSON],
                'rel': 'http://www.opengis.net/def/rel/ogc/1.0/execute',
                'href': f'{process_url}/execution?f={F_JSON}',
                'title': l10n.translate('Execution for this process as JSON', request.locale),  # noqa
                'hreflang': api.default_locale
            }
            p2['links'].append(link)

            processes.append(p2)

    if process is not None:
        response = processes[0]
    else:
        process_url = f"{api.base_url}/processes"
        response = {
            'processes': processes,
            'links': [{
                'type': FORMAT_TYPES[F_JSON],
                'rel': request.get_linkrel(F_JSON),
                'title': l10n.translate('This document as JSON', request.locale),  # noqa
                'href': f'{process_url}?f={F_JSON}'
            }, {
                'type': FORMAT_TYPES[F_JSONLD],
                'rel': request.get_linkrel(F_JSONLD),
                'title': l10n.translate('This document as RDF (JSON-LD)', request.locale),  # noqa
                'href': f'{process_url}?f={F_JSONLD}'
            }, {
                'type': FORMAT_TYPES[F_HTML],
                'rel': request.get_linkrel(F_HTML),
                'title': l10n.translate('This document as HTML', request.locale),  # noqa
                'href': f'{process_url}?f={F_HTML}'
            }]
        }

    if request.format == F_HTML:  # render
        if process is not None:
            tpl_config = api.get_dataset_templates(process)
            response = render_j2_template(api.tpl_config, tpl_config,
                                          'processes/process.html',
                                          response, request.locale)
        else:
            response = render_j2_template(
                api.tpl_config, api.config['server']['templates'],
                'processes/index.html', response, request.locale)

        return headers, HTTPStatus.OK, response

    return headers, HTTPStatus.OK, to_json(response, api.pretty_print)


# TODO: get_jobs doesn't have tests
def get_jobs(api: API, request: APIRequest,
             job_id: str | None = None) -> Tuple[dict, int, str]:
    """
    Get process jobs

    :param request: A request object
    :param job_id: id of job

    :returns: tuple of headers, status code, content
    """

    headers = request.get_response_headers(SYSTEM_LOCALE,
                                           **api.api_headers)
    LOGGER.debug('Processing limit parameter')
    try:
        limit = evaluate_limit(request.params.get('limit'),
                               api.config['server'].get('limits', {}),
                               {})
    except ValueError as err:
        return api.get_exception(
            HTTPStatus.BAD_REQUEST, headers, request.format,
            'InvalidParameterValue', str(err))

    LOGGER.debug('Processing offset parameter')
    try:
        offset = int(request.params.get('offset'))
        if offset < 0:
            msg = 'offset value should be positive or zero'
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'InvalidParameterValue', msg)
    except TypeError as err:
        LOGGER.warning(err)
        offset = 0
    except ValueError:
        msg = 'offset value should be an integer'
        return api.get_exception(
            HTTPStatus.BAD_REQUEST, headers, request.format,
            'InvalidParameterValue', msg)

    if job_id is None:
        jobs_data = api.manager.get_jobs(limit=limit, offset=offset)
        # TODO: For pagination to work, the provider has to do the sorting.
        #       Here we do sort again in case the provider doesn't support
        #       pagination yet and always returns all jobs.
        jobs = sorted(jobs_data['jobs'],
                      key=lambda k: k['started'],
                      reverse=True)
        numberMatched = jobs_data['numberMatched']

    else:
        try:
            jobs = [api.manager.get_job(job_id)]
        except JobNotFoundError:
            return api.get_exception(
                HTTPStatus.NOT_FOUND, headers, request.format,
                'InvalidParameterValue', job_id)
        numberMatched = 1

    serialized_jobs = {
        'jobs': [],
        'links': [{
            'href': f"{api.base_url}/jobs?f={F_HTML}",
            'rel': request.get_linkrel(F_HTML),
            'type': FORMAT_TYPES[F_HTML],
            'title': l10n.translate('Jobs list as HTML', request.locale)
        }, {
            'href': f"{api.base_url}/jobs?f={F_JSON}",
            'rel': request.get_linkrel(F_JSON),
            'type': FORMAT_TYPES[F_JSON],
            'title': l10n.translate('Jobs list as JSON', request.locale)
        }]
    }
    for job_ in jobs:
        job2 = {
            'type': 'process',
            'processID': job_['process_id'],
            'jobID': job_['identifier'],
            'status': job_['status'],
            'message': job_['message'],
            'progress': job_['progress'],
            'parameters': job_.get('parameters'),
            'created': job_['created'],
            'started': job_['started'],
            'finished': job_['finished'],
            'updated': job_['updated']
        }

        # TODO: translate
        if JobStatus[job_['status']] in (
           JobStatus.successful, JobStatus.running, JobStatus.accepted):

            job_result_url = f"{api.base_url}/jobs/{job_['identifier']}/results"  # noqa

            job2['links'] = [{
                'href': f'{job_result_url}?f={F_HTML}',
                'rel': 'http://www.opengis.net/def/rel/ogc/1.0/results',
                'type': FORMAT_TYPES[F_HTML],
                'title': l10n.translate(f'Results of job as HTML', request.locale),  # noqa
            }, {
                'href': f'{job_result_url}?f={F_JSON}',
                'rel': 'http://www.opengis.net/def/rel/ogc/1.0/results',
                'type': FORMAT_TYPES[F_JSON],
                'title': l10n.translate(f'Results of job as JSON', request.locale),  # noqa
            }]

            if job_['mimetype'] not in (FORMAT_TYPES[F_JSON],
                                        FORMAT_TYPES[F_HTML]):

                job2['links'].append({
                    'href': job_result_url,
                    'rel': 'http://www.opengis.net/def/rel/ogc/1.0/results',  # noqa
                    'type': job_['mimetype'],
                    'title': f"Results of job {job_id} as {job_['mimetype']}"  # noqa
                })

        serialized_jobs['jobs'].append(job2)

    serialized_query_params = ''
    for k, v in request.params.items():
        if k not in ('f', 'offset'):
            serialized_query_params += '&'
            serialized_query_params += urllib.parse.quote(k, safe='')
            serialized_query_params += '='
            serialized_query_params += urllib.parse.quote(str(v), safe=',')

    uri = f'{api.base_url}/jobs'

    if offset > 0:
        prev = max(0, offset - limit)
        serialized_jobs['links'].append(
            {
                'href': f'{uri}?offset={prev}{serialized_query_params}',
                'type': FORMAT_TYPES[F_JSON],
                'rel': 'prev',
                'title': l10n.translate('Items (prev)', request.locale),
            })

    next_link = False

    if numberMatched > (limit + offset):
        next_link = True
    elif len(jobs) == limit:
        next_link = True

    if next_link:
        next_ = offset + limit
        next_href = f'{uri}?offset={next_}{serialized_query_params}'
        serialized_jobs['links'].append(
            {
                'href': next_href,
                'rel': 'next',
                'type': FORMAT_TYPES[F_JSON],
                'title': l10n.translate('Items (next)', request.locale),
            })

    if job_id is None:
        j2_template = 'jobs/index.html'
    else:
        serialized_jobs = serialized_jobs['jobs'][0]
        j2_template = 'jobs/job.html'

    if request.format == F_HTML:
        data = {
            'jobs': serialized_jobs,
            'offset': offset,
            'now': datetime.now(timezone.utc).strftime(DATETIME_FORMAT)
        }
        response = render_j2_template(
            api.tpl_config, api.config['server']['templates'], j2_template,
            data, request.locale)

        return headers, HTTPStatus.OK, response

    return headers, HTTPStatus.OK, to_json(serialized_jobs,
                                           api.pretty_print)


def execute_process(api: API, request: APIRequest,
                    process_id: str) -> Tuple[dict, int, str]:
    """
    Execute process

    :param request: A request object
    :param process_id: id of process

    :returns: tuple of headers, status code, content
    """

    # Responses are always in US English only
    headers = request.get_response_headers(SYSTEM_LOCALE,
                                           **api.api_headers)
    if process_id not in api.manager.processes:
        msg = 'identifier not found'
        return api.get_exception(
            HTTPStatus.NOT_FOUND, headers,
            request.format, 'NoSuchProcess', msg)

    data = request.data
    if not data:
        # TODO not all processes require input, e.g. time-dependent or
        #      random value generators
        msg = 'missing request data'
        return api.get_exception(
            HTTPStatus.BAD_REQUEST, headers, request.format,
            'MissingParameterValue', msg)

    try:
        # Parse bytes data, if applicable
        data = data.decode()
        LOGGER.debug(data)
    except (UnicodeDecodeError, AttributeError):
        pass

    try:
        data = json.loads(data)
    except (json.decoder.JSONDecodeError, TypeError):
        # Input does not appear to be valid JSON
        msg = 'invalid request data'
        return api.get_exception(
            HTTPStatus.BAD_REQUEST, headers, request.format,
            'InvalidParameterValue', msg)

    data_dict = data.get('inputs', {})
    LOGGER.debug(data_dict)

    requested_outputs = data.get('outputs')
    LOGGER.debug(f'outputs: {requested_outputs}')

    requested_response = data.get('response', 'raw')

    subscriber = None
    subscriber_dict = data.get('subscriber')
    if subscriber_dict:
        try:
            success_uri = subscriber_dict['successUri']
        except KeyError:
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'MissingParameterValue', 'Missing successUri')
        else:
            subscriber = Subscriber(
                # NOTE: successUri is mandatory according to the standard
                success_uri=success_uri,
                in_progress_uri=subscriber_dict.get('inProgressUri'),
                failed_uri=subscriber_dict.get('failedUri'),
            )

    try:
        execution_mode = RequestedProcessExecutionMode(
            request.headers.get('Prefer', request.headers.get('prefer'))
        )
    except ValueError:
        execution_mode = None
    try:
        LOGGER.debug('Executing process')
        result = api.manager.execute_process(
            process_id, data_dict, execution_mode=execution_mode,
            requested_outputs=requested_outputs,
            subscriber=subscriber,
            requested_response=requested_response)
        job_id, mime_type, outputs, status, additional_headers = result
        headers.update(additional_headers or {})

        if api.manager.is_async:
            headers['Location'] = f'{api.base_url}/jobs/{job_id}'

    except ProcessorExecuteError as err:
        return api.get_exception(
            err.http_status_code, headers,
            request.format, err.ogc_exception_code, err.message)

    response = {}
    if status == JobStatus.failed:
        response = outputs

    if requested_response == 'raw':
        headers['Content-Type'] = mime_type
        response = outputs
    elif status not in (JobStatus.failed, JobStatus.accepted):
        response = outputs

    if status == JobStatus.accepted:
        http_status = HTTPStatus.CREATED
    elif status == JobStatus.failed:
        http_status = HTTPStatus.BAD_REQUEST
    else:
        http_status = HTTPStatus.OK

    if mime_type == 'application/json':
        if requested_response == 'document':
            pretty_print_ = api.pretty_print
        else:  # raw
            pretty_print_ = False
        response2 = to_json(response, pretty_print_)
    else:
        response2 = response

    if (headers.get('Preference-Applied', '') == RequestedProcessExecutionMode.respond_async.value):  # noqa
        LOGGER.debug('Asynchronous mode detected, returning statusInfo')
        response2 = {
            'jobID': job_id,
            'type': 'process',
            'status': status.value
        }
        response2 = to_json(response2, pretty_print_)

    if api.pubsub_client is not None:
        LOGGER.debug('Publishing message')
        try:
            publish_message(api.pubsub_client, api.base_url, 'process',
                            process_id, job_id, response2)
        except Exception as err:
            msg = f'Could not publish message {err}'
            LOGGER.warning(msg)

    return headers, http_status, response2


def get_job_result(api: API, request: APIRequest,
                   job_id: str) -> Tuple[dict, int, str]:
    """
    Get result of job (instance of a process)

    :param request: A request object
    :param job_id: ID of job

    :returns: tuple of headers, status code, content
    """

    headers = request.get_response_headers(SYSTEM_LOCALE,
                                           **api.api_headers)
    try:
        job = api.manager.get_job(job_id)
    except JobNotFoundError:
        return api.get_exception(
            HTTPStatus.NOT_FOUND, headers,
            request.format, 'NoSuchJob', job_id
        )

    status = JobStatus[job['status']]

    if status == JobStatus.running:
        msg = 'job still running'
        return api.get_exception(
            HTTPStatus.NOT_FOUND, headers,
            request.format, 'ResultNotReady', msg)

    elif status == JobStatus.accepted:
        # NOTE: this case is not mentioned in the specification
        msg = 'job accepted but not yet running'
        return api.get_exception(
            HTTPStatus.NOT_FOUND, headers,
            request.format, 'ResultNotReady', msg)

    elif status == JobStatus.failed:
        msg = 'job failed'
        return api.get_exception(
            HTTPStatus.BAD_REQUEST, headers, request.format,
            'InvalidParameterValue', msg)

    try:
        mimetype, job_output = api.manager.get_job_result(job_id)
    except JobResultNotFoundError:
        return api.get_exception(
            HTTPStatus.INTERNAL_SERVER_ERROR, headers,
            request.format, 'JobResultNotFound', job_id
        )

    if mimetype not in (None, FORMAT_TYPES[F_JSON]):
        headers['Content-Type'] = mimetype
        content = job_output
    else:
        if request.format == F_JSON:
            content = json.dumps(job_output, sort_keys=True, indent=4,
                                 default=json_serial)
        else:
            # HTML
            headers['Content-Type'] = "text/html"
            data = {
                'job': {'id': job_id},
                'result': job_output
            }
            content = render_j2_template(
                api.config, api.config['server']['templates'],
                'jobs/results/index.html', data, request.locale)

    return headers, HTTPStatus.OK, content


def delete_job(api: API, request: APIRequest,
               job_id: str) -> Tuple[dict, int, str]:
    """
    Delete a process job

    :param job_id: job identifier

    :returns: tuple of headers, status code, content
    """

    response_headers = request.get_response_headers(
        SYSTEM_LOCALE, **api.api_headers)
    try:
        success = api.manager.delete_job(job_id)
    except JobNotFoundError:
        return api.get_exception(
            HTTPStatus.NOT_FOUND, response_headers, request.format,
            'NoSuchJob', job_id
        )
    else:
        if success:
            http_status = HTTPStatus.OK
            jobs_url = f"{api.base_url}/jobs"

            response = {
                'jobID': job_id,
                'status': JobStatus.dismissed.value,
                'message': 'Job dismissed',
                'progress': 100,
                'links': [{
                    'href': jobs_url,
                    'rel': 'up',
                    'type': FORMAT_TYPES[F_JSON],
                    'title': l10n.translate('The job list for the current process', request.locale)  # noqa
                }]
            }
        else:
            return api.get_exception(
                HTTPStatus.INTERNAL_SERVER_ERROR, response_headers,
                request.format, 'InternalError', job_id
            )
    LOGGER.info(response)
    # TODO: this response does not have any headers
    return {}, http_status, to_json(response, api.pretty_print)


def get_oas_30(cfg: dict, locale: str
               ) -> tuple[list[dict[str, str]], dict[str, dict]]:  # noqa
    """
    Get OpenAPI fragments

    :param cfg: `dict` of configuration
    :param locale: `str` of locale

    :returns: `tuple` of `list` of tag objects, and `dict` of path objects
    """

    from pygeoapi.openapi import OPENAPI_YAML

    LOGGER.debug('setting up processes endpoints')

    oas = {'tags': []}

    paths = {}

    process_manager = get_manager(cfg)

    if len(process_manager.processes) > 0:
        paths['/processes'] = {
            'get': {
                'summary': 'Processes',
                'description': 'Processes',
                'tags': ['server'],
                'operationId': 'getProcesses',
                'parameters': [
                    {'$ref': '#/components/parameters/f'}
                ],
                'responses': {
                    '200': {'$ref': f"{OPENAPI_YAML['oapip']}/responses/ProcessList.yaml"},  # noqa
                    'default': {'$ref': '#/components/responses/default'}
                }
            }
        }

    LOGGER.debug('setting up processes')

    for k, v in process_manager.processes.items():
        if k.startswith('_'):
            LOGGER.debug(f'Skipping hidden layer: {k}')
            continue
        name = l10n.translate(k, locale)
        p = process_manager.get_processor(k)
        md_desc = l10n.translate(p.metadata['description'], locale)
        process_name_path = f'/processes/{name}'
        tag = {
            'name': name,
            'description': md_desc,
            'externalDocs': {}
        }
        for link in p.metadata.get('links', []):
            if link.get('rel', '') == 'information':
                translated_link = l10n.translate(link, locale)
                tag['externalDocs']['description'] = translated_link[
                    'rel']
                tag['externalDocs']['url'] = translated_link['href']
                break
        if len(tag['externalDocs']) == 0:
            del tag['externalDocs']

        oas['tags'].append(tag)

        paths[process_name_path] = {
            'get': {
                'summary': 'Get process metadata',
                'description': md_desc,
                'tags': [name],
                'operationId': f'describe{name.capitalize()}Process',
                'parameters': [
                    {'$ref': '#/components/parameters/f'}
                ],
                'responses': {
                    '200': {'$ref': '#/components/responses/200'},
                    'default': {'$ref': '#/components/responses/default'}
                }
            }
        }

        paths[f'{process_name_path}/execution'] = {
            'post': {
                'summary': f"Process {l10n.translate(p.metadata['title'], locale)} execution",  # noqa
                'description': md_desc,
                'tags': [name],
                'operationId': f'execute{name.capitalize()}Job',
                'parameters': [{
                    'in': 'header',
                    'name': 'Prefer',
                    'required': False,
                    'description': 'Indicates client preferences, including whether the client is capable of asynchronous processing.',  # noqa
                    'schema': {
                        'type': 'string',
                        'enum': []
                    }
                }],
                'responses': {
                    '200': {'$ref': '#/components/responses/200'},
                    '201': {'$ref': f"{OPENAPI_YAML['oapip']}/responses/ExecuteAsync.yaml"},  # noqa
                    '404': {'$ref': f"{OPENAPI_YAML['oapip']}/responses/NotFound.yaml"},  # noqa
                    '500': {'$ref': f"{OPENAPI_YAML['oapip']}/responses/ServerError.yaml"},  # noqa
                    'default': {'$ref': '#/components/responses/default'}
                },
                'requestBody': {
                    'description': 'Mandatory execute request JSON',
                    'required': True,
                    'content': {
                        'application/json': {
                            'schema': {
                                '$ref': f"{OPENAPI_YAML['oapip']}/schemas/execute.yaml"  # noqa
                            }
                        }
                    }
                }
            }
        }

        jco = p.metadata.get('jobControlOptions', ['sync-execute'])
        if 'sync-execute' in jco:
            paths[f'{process_name_path}/execution']['post']['parameters'][0]['schema']['enum'].append('respond-sync')  # noqa
        if 'async-execute' in jco:
            paths[f'{process_name_path}/execution']['post']['parameters'][0]['schema']['enum'].append('respond-async')  # noqa

        try:
            first_key = list(p.metadata['outputs'])[0]
            p_output = p.metadata['outputs'][first_key]

            if p_output.get('schema') is not None:
                LOGGER.debug('Adding output schema')
                content_media_type = p_output['schema'].pop('contentMediaType', 'application/json')  # noqa
                paths[f'{process_name_path}/execution']['post']['responses']['200'] = {  # noqa
                    'description': 'Process output schema',
                    'content': {
                        content_media_type: {
                            'schema': p_output['schema']
                        }
                    }
                }
        except (IndexError, KeyError):
            LOGGER.debug('No output defined')

        if 'example' in p.metadata:
            paths[f'{process_name_path}/execution']['post']['requestBody']['content']['application/json']['example'] = p.metadata['example']  # noqa

    name_in_path = {
        'name': 'jobId',
        'in': 'path',
        'description': 'job identifier',
        'required': True,
        'schema': {
            'type': 'string'
        }
    }

    paths['/jobs'] = {
        'get': {
            'summary': 'Retrieve jobs list',
            'description': 'Retrieve a list of jobs',
            'tags': ['jobs'],
            'operationId': 'getJobs',
            'responses': {
                '200': {'$ref': '#/components/responses/200'},
                '404': {'$ref': f"{OPENAPI_YAML['oapip']}/responses/NotFound.yaml"},  # noqa
                'default': {'$ref': '#/components/responses/default'}
            }
        }
    }

    paths['/jobs/{jobId}'] = {
        'get': {
            'summary': 'Retrieve job details',
            'description': 'Retrieve job details',
            'tags': ['jobs'],
            'parameters': [
                name_in_path,
                {'$ref': '#/components/parameters/f'}
            ],
            'operationId': 'getJob',
            'responses': {
                '200': {'$ref': '#/components/responses/200'},
                '404': {'$ref': f"{OPENAPI_YAML['oapip']}/responses/NotFound.yaml"},  # noqa
                'default': {'$ref': '#/components/responses/default'}
            }
        },
        'delete': {
            'summary': 'Cancel / delete job',
            'description': 'Cancel / delete job',
            'tags': ['jobs'],
            'parameters': [
                name_in_path
            ],
            'operationId': 'deleteJob',
            'responses': {
                '204': {'$ref': '#/components/responses/204'},
                '404': {'$ref': f"{OPENAPI_YAML['oapip']}/responses/NotFound.yaml"},  # noqa
                'default': {'$ref': '#/components/responses/default'}
            }
        },
    }

    paths['/jobs/{jobId}/results'] = {
        'get': {
            'summary': 'Retrieve job results',
            'description': 'Retrieve job results',
            'tags': ['jobs'],
            'parameters': [
                name_in_path,
                {'$ref': '#/components/parameters/f'}
            ],
            'operationId': 'getJobResults',
            'responses': {
                '200': {'$ref': '#/components/responses/200'},
                '404': {'$ref': f"{OPENAPI_YAML['oapip']}/responses/NotFound.yaml"},  # noqa
                'default': {'$ref': '#/components/responses/default'}
            }
        }
    }

    return [{'name': 'processes'}, {'name': 'jobs'}], {'paths': paths}



# =================================================================
#
# Authors: Leo Ghignone <leo.ghignone@gmail.com>
#
# Copyright (c) 2024 Leo Ghignone
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from itertools import chain
import json
import logging

from dateutil.parser import isoparse
import geopandas as gpd
import pyarrow
import pyarrow.compute as pc
import pyarrow.dataset
import s3fs

from pygeoapi.crs import crs_transform
from pygeoapi.provider.base import (
    BaseProvider,
    ProviderConnectionError,
    ProviderGenericError,
    ProviderItemNotFoundError,
    ProviderQueryError,
)

LOGGER = logging.getLogger(__name__)


def arrow_to_pandas_type(arrow_type):
    pd_type = arrow_type.to_pandas_dtype()
    try:
        # Needed for specific types such as dtype('<M8[ns]')
        pd_type = pd_type.type
    except AttributeError:
        pd_type = pd_type
    return pd_type


class ParquetProvider(BaseProvider):
    def __init__(self, provider_def):
        """
        Initialize object

        # Typical ParquetProvider YAML config:

        provider:
            name: Parquet
            data:
                source: s3://example.com/parquet_directory/

            id_field: gml_id


        :param provider_def: provider definition

        :returns: pygeoapi.provider.parquet.ParquetProvider
        """

        super().__init__(provider_def)

        # Source url is required
        self.source = self.data.get('source')
        if not self.source:
            msg = "Need explicit 'source' attr " \
                    "in data field of provider config"
            LOGGER.error(msg)
            raise Exception(msg)

        # Manage AWS S3 sources
        if self.source.startswith('s3'):
            self.source = self.source.split('://', 1)[1]
            self.fs = s3fs.S3FileSystem(default_cache_type='none')
        else:
            self.fs = None

        # Build pyarrow dataset pointing to the data
        self.ds = pyarrow.dataset.dataset(self.source, filesystem=self.fs)

        LOGGER.debug('Grabbing field information')
        self.get_fields()  # Must be set to visualise queryables

        # Column names for bounding box data.
        if None in [self.x_field, self.y_field]:
            self.has_geometry = False
        else:
            self.has_geometry = True
            if isinstance(self.x_field, str):
                self.minx = self.x_field
                self.maxx = self.x_field
            else:
                self.minx, self.maxx = self.x_field

            if isinstance(self.y_field, str):
                self.miny = self.y_field
                self.maxy = self.y_field
            else:
                self.miny, self.maxy = self.y_field
            self.bb = [self.minx, self.miny, self.maxx, self.maxy]

            # Get the CRS of the data
            geo_metadata = json.loads(self.ds.schema.metadata[b'geo'])
            geom_column = geo_metadata['primary_column']
            # if the CRS is not set default to EPSG:4326, per geoparquet spec
            self.crs = (geo_metadata['columns'][geom_column].get('crs')
                        or 'OGC:CRS84')

    def _read_parquet(self, return_scanner=False, **kwargs):
        """
        Scan a Parquet dataset with the given arguments

        :returns: generator of RecordBatch with the queried values
        """
        scanner = pyarrow.dataset.Scanner.from_dataset(self.ds, **kwargs)
        batches = scanner.to_batches()
        if return_scanner:
            return batches, scanner
        else:
            return batches

    def get_fields(self):
        """
        Get provider field information (names, types)

        :returns: dict of fields
        """

        if not self._fields:

            for field_name, field_type in zip(self.ds.schema.names,
                                              self.ds.schema.types):
                # Geometry is managed as a special case by pygeoapi
                if field_name == 'geometry':
                    continue

                field_type = str(field_type)
                converted_type = None
                converted_format = None
                if field_type.startswith(('int', 'uint')):
                    converted_type = 'integer'
                    converted_format = field_type
                elif field_type == 'double' or field_type.startswith('float'):
                    converted_type = 'number'
                    converted_format = field_type
                elif field_type == 'string':
                    converted_type = 'string'
                elif field_type == 'bool':
                    converted_type = 'boolean'
                elif field_type.startswith('timestamp'):
                    converted_type = 'string'
                    converted_format = 'date-time'
                else:
                    LOGGER.error(f'Unsupported field type {field_type}')

                if converted_format is None:
                    self._fields[field_name] = {'type': converted_type}
                else:
                    self._fields[field_name] = {
                        'type': converted_type,
                        'format': converted_format,
                    }

        return self._fields

    @crs_transform
    def query(
        self,
        offset=0,
        limit=10,
        resulttype='results',
        bbox=[],
        datetime_=None,
        properties=[],
        select_properties=[],
        skip_geometry=False,
        q=None,
        **kwargs,
    ):
        """
        Query Parquet source

        :param offset: starting record to return (default 0)
        :param limit: number of records to return (default 10)
        :param resulttype: return results or hit limit (default results)
        :param bbox: bounding box [minx,miny,maxx,maxy]
        :param datetime_: temporal (datestamp or extent) following ISO-8601
        :param properties: list of tuples (field, comparison, value)
        :param select_properties: list of property names
        :param skip_geometry: bool of whether to skip geometry (default False)

        :returns: dict of 0..n GeoJSON features
        """
        result = None
        try:
            filter = pc.scalar(True)
            if bbox:
                if self.has_geometry is False:
                    msg = (
                        'Dataset does not have a geometry field, '
                        'querying by bbox is not supported.'
                    )
                    raise ProviderQueryError(msg)
                LOGGER.debug('processing bbox parameter')
                if any(b is None for b in bbox):
                    msg = 'Dataset does not support bbox filtering'
                    raise ProviderQueryError(msg)

                minx, miny, maxx, maxy = [float(b) for b in bbox]
                filter = (
                    (pc.field(self.minx) > pc.scalar(minx))
                    & (pc.field(self.miny) > pc.scalar(miny))
                    & (pc.field(self.maxx) < pc.scalar(maxx))
                    & (pc.field(self.maxy) < pc.scalar(maxy))
                )

            if datetime_ is not None:
                if self.time_field is None:
                    msg = (
                        'Dataset does not have a time field, '
                        'querying by datetime is not supported.'
                    )
                    raise ProviderQueryError(msg)
                timefield = pc.field(self.time_field)
                if '/' in datetime_:
                    begin, end = datetime_.split('/')
                    if begin != '..':
                        begin = isoparse(begin)
                        filter = filter & (timefield >= begin)
                    if end != '..':
                        end = isoparse(end)
                        filter = filter & (timefield <= end)
                else:
                    target_time = isoparse(datetime_)
                    filter = filter & (timefield == target_time)

            if properties:
                LOGGER.debug('processing properties')
                for name, value in properties:
                    field = self.ds.schema.field(name)
                    pd_type = arrow_to_pandas_type(field.type)
                    expr = pc.field(name) == pc.scalar(pd_type(value))

                    filter = filter & expr

            if len(select_properties) == 0:
                select_properties = self.ds.schema.names
            else:  # Load id and geometry together with any specified columns
                if self.has_geometry and 'geometry' not in select_properties:
                    select_properties.append('geometry')
                if self.id_field not in select_properties:
                    select_properties.insert(0, self.id_field)

            if skip_geometry:
                select_properties.remove('geometry')

            # Make response based on resulttype specified
            if resulttype == 'hits':
                LOGGER.debug('hits only specified')
                result = self._response_feature_hits(filter)
            elif resulttype == 'results':
                LOGGER.debug('results specified')
                result = self._response_feature_collection(
                    filter, offset, limit, columns=select_properties
                )
            else:
                LOGGER.error(f'Invalid resulttype: {resulttype}')

        except RuntimeError as err:
            LOGGER.error(err)
            raise ProviderQueryError(err)
        except ProviderConnectionError as err:
            LOGGER.error(err)
            raise ProviderConnectionError(err)
        except Exception as err:
            LOGGER.error(err)
            raise ProviderGenericError(err)

        return result

    @crs_transform
    def get(self, identifier, **kwargs):
        """
        Get Feature by id

        :param identifier: feature id

        :returns: a single feature
        """
        result = None
        try:
            LOGGER.debug(f'Fetching identifier {identifier}')
            id_type = arrow_to_pandas_type(
                self.ds.schema.field(self.id_field).type)
            batches = self._read_parquet(
                filter=(
                    pc.field(self.id_field) == pc.scalar(id_type(identifier))
                )
            )

            for batch in batches:
                if batch.num_rows > 0:
                    assert (
                        batch.num_rows == 1
                    ), f'Multiple items found with ID {identifier}'
                    row = batch.to_pandas()
                    break
            else:
                raise ProviderItemNotFoundError(f'ID {identifier} not found')

            if self.has_geometry:
                geom = gpd.GeoSeries.from_wkb(row['geometry'], crs=self.crs)
            else:
                geom = [None]
            gdf = gpd.GeoDataFrame(row, geometry=geom)
            LOGGER.debug('results computed')

            # Grab the collection from geopandas geo_interface
            result = gdf.__geo_interface__['features'][0]

        except RuntimeError as err:
            LOGGER.error(err)
            raise ProviderQueryError(err)
        except ProviderConnectionError as err:
            LOGGER.error(err)
            raise ProviderConnectionError(err)
        except ProviderItemNotFoundError as err:
            LOGGER.error(err)
            raise ProviderItemNotFoundError(err)
        except Exception as err:
            LOGGER.error(err)
            raise ProviderGenericError(err)

        return result

    def __repr__(self):
        return f'<ParquetProvider> {self.data}'

    def _response_feature_collection(self, filter, offset, limit,
                                     columns=None):
        """
        Assembles output from query as
        GeoJSON FeatureCollection structure.

        :returns: GeoJSON FeatureCollection
        """

        LOGGER.debug(f'offset:{offset}, limit:{limit}')

        try:
            batches, scanner = self._read_parquet(
                filter=filter, columns=columns, return_scanner=True
            )

            # Discard batches until offset is reached
            counted = 0
            for batch in batches:
                if counted + batch.num_rows > offset:
                    # Slice current batch to start from the requested row
                    batch = batch.slice(offset=offset - counted)
                    # Build a new generator yielding the current batch
                    # and all following ones

                    batches = chain([batch], batches)
                    break
                else:
                    counted += batch.num_rows

            # batches is a generator, it will now be either fully spent
            # or set to the new generator starting from offset

            # Get the next `limit+1` rows
            # The extra row is used to check if a "next" link is needed
            # (when numberMatched > offset + limit)
            batches_list = []
            read = 0

            for batch in batches:
                read += batch.num_rows
                if read > limit:
                    batches_list.append(batch.slice(0, limit + 1))
                    break
                else:
                    batches_list.append(batch)

            # Passing schema from scanner in case no rows are returned
            table = pyarrow.Table.from_batches(
                batches_list, schema=scanner.projected_schema
            )

            rp = table.to_pandas()

            number_matched = offset + len(rp)

            # Remove the extra row
            if len(rp) > limit:
                rp = rp.iloc[:-1]

            if 'geometry' not in rp.columns:
                # We need a null geometry column to create a GeoDataFrame
                rp['geometry'] = None
                geom = gpd.GeoSeries.from_wkb(rp['geometry'])
            else:
                geom = gpd.GeoSeries.from_wkb(rp['geometry'], crs=self.crs)

            gdf = gpd.GeoDataFrame(rp, geometry=geom)
            LOGGER.debug('results computed')
            result = gdf.__geo_interface__

            # Add numberMatched to generate "next" link
            result['numberMatched'] = number_matched

            return result

        except RuntimeError as error:
            LOGGER.error(error)
            raise error

    def _response_feature_hits(self, filter):
        """
        Assembles GeoJSON hits from row count

        :returns: GeoJSON FeaturesCollection
        """

        try:
            scanner = pyarrow.dataset.Scanner.from_dataset(self.ds,
                                                           filter=filter)
            return {
                'type': 'FeatureCollection',
                'numberMatched': scanner.count_rows(),
                'features': [],
            }
        except Exception as error:
            LOGGER.error(error)
            raise error



# =================================================================
#
# Authors: Prajwal Amaravati <prajwal.s@satsure.co>
#          Tanvi Prasad <tanvi.prasad@cdpg.org.in>
#          Bryan Robert <bryan.robert@cdpg.org.in>
#          Benjamin Webb <bwebb@lincolninst.edu>
#
# Copyright (c) 2025 Prajwal Amaravati
# Copyright (c) 2025 Tanvi Prasad
# Copyright (c) 2025 Bryan Robert
# Copyright (c) 2025 Benjamin Webb
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import logging

from geoalchemy2.functions import (ST_TileEnvelope, ST_Transform, ST_AsMVTGeom,
                                   ST_AsMVT, ST_CurveToLine, ST_MakeEnvelope)

from sqlalchemy.sql import select
from sqlalchemy.orm import Session

from pygeoapi.crs import get_crs
from pygeoapi.models.provider.base import (
    TileSetMetadata, TileMatrixSetEnum, LinkType)
from pygeoapi.provider.base import ProviderConnectionError
from pygeoapi.provider.base_mvt import BaseMVTProvider
from pygeoapi.provider.sql import PostgreSQLProvider
from pygeoapi.provider.tile import ProviderTileNotFoundError
from pygeoapi.util import url_join

LOGGER = logging.getLogger(__name__)


class MVTPostgreSQLProvider(BaseMVTProvider, PostgreSQLProvider):
    """
    MVT PostgreSQL Provider
    Provider for serving tiles rendered on-the-fly from
    feature tables in PostgreSQL
    """

    def __init__(self, provider_def):
        """
        Initialize object

        :param provider_def: provider definition

        :returns: pygeoapi.provider.MVT.MVTPostgreSQLProvider
        """
        PostgreSQLProvider.__init__(self, provider_def)
        BaseMVTProvider.__init__(self, provider_def)

    def get_fields(self):
        """
        Get Postgres columns

        :returns: `list` of columns
        """
        if not self._fields:
            for column in self.table_model.__table__.columns:
                LOGGER.debug(f'Testing {column.name}')
                if column.name == self.geom:
                    continue

                self._fields[str(column.name)] = (
                    column.label('id')
                    if column.name == self.id_field else
                    column
                )

        return self._fields

    def get_layer(self):
        """
        Use table name as layer name

        :returns: `str` of layer name
        """
        return self.table

    def get_tiling_schemes(self):
        """
        Only WebMercatorQuad and WorldCRS84Quad tiling schemes
        are supported for PostgreSQL
        """

        return [
            TileMatrixSetEnum.WEBMERCATORQUAD.value,
            TileMatrixSetEnum.WORLDCRS84QUAD.value
        ]

    def get_tiles_service(self, baseurl=None, servicepath=None,
                          dirpath=None, tile_type=None):
        """
        Gets mvt service description

        :param baseurl: base URL of endpoint
        :param servicepath: base path of URL
        :param dirpath: directory basepath (equivalent of URL)
        :param tile_type: tile format type

        :returns: `dict` of item tile service
        """

        BaseMVTProvider.get_tiles_service(self,
                                          baseurl, servicepath,
                                          dirpath, tile_type)

        self._service_url = servicepath
        return self.get_tms_links()

    def get_tiles(self, layer='default', tileset=None,
                  z=None, y=None, x=None, format_=None):
        """
        Gets tile

        :param layer: mvt tile layer
        :param tileset: mvt tileset
        :param z: z index
        :param y: y index
        :param x: x index
        :param format_: tile format

        :returns: an encoded mvt tile
        """
        z, y, x = map(int, [z, y, x])

        [tileset_schema] = [
            schema for schema in self.get_tiling_schemes()
            if tileset == schema.tileMatrixSet
        ]
        if not self.is_in_limits(tileset_schema, z, x, y):
            LOGGER.warning(f'Tile {z}/{x}/{y} not found')
            raise ProviderTileNotFoundError

        envelope = self.get_envelope(z, y, x, tileset)
        geom_column = getattr(self.table_model, self.geom)
        geom_filter = geom_column.intersects(
            ST_Transform(envelope, self.storage_crs.to_string())
        )

        out_srid = get_crs(tileset_schema.crs).to_string()
        mvtgeom = (
            ST_AsMVTGeom(
                ST_Transform(ST_CurveToLine(geom_column), out_srid),
                ST_Transform(envelope, out_srid))
            .label('mvtgeom')
        )

        mvtrow = (
            select(mvtgeom, *self.fields.values())
            .filter(geom_filter)
            .cte('mvtrow')
        )

        mvtquery = select(
            ST_AsMVT(mvtrow.table_valued(), layer)
        )

        with Session(self._engine) as session:
            result = bytes(
                session.execute(mvtquery).scalar()
            ) or None

        return result

    def get_html_metadata(self, dataset, server_url, layer, tileset,
                          title, description, keywords, **kwargs):

        service_url = url_join(
            server_url,
            f'collections/{dataset}/tiles/{tileset}',
            '{tileMatrix}/{tileRow}/{tileCol}?f=mvt')
        metadata_url = url_join(
            server_url,
            f'collections/{dataset}/tiles/{tileset}/metadata')

        metadata = dict()
        metadata['id'] = dataset
        metadata['title'] = title
        metadata['tileset'] = tileset
        metadata['collections_path'] = service_url
        metadata['json_url'] = f'{metadata_url}?f=json'

        return metadata

    def get_default_metadata(self, dataset, server_url, layer, tileset,
                             title, description, keywords, **kwargs):

        service_url = url_join(
            server_url,
            f'collections/{dataset}/tiles/{tileset}',
            '{tileMatrix}/{tileRow}/{tileCol}?f=mvt'
            )

        tiling_schemes = self.get_tiling_schemes()
        # Default values
        tileMatrixSetURI = tiling_schemes[0].tileMatrixSetURI
        crs = tiling_schemes[0].crs
        # Checking the selected matrix in configured tiling_schemes
        for schema in tiling_schemes:
            if (schema.tileMatrixSet == tileset):
                crs = schema.crs
                tileMatrixSetURI = schema.tileMatrixSetURI

                tiling_scheme_url = url_join(
                    server_url, 'TileMatrixSets', schema.tileMatrixSet)
                tiling_scheme_url_type = 'application/json'
                tiling_scheme_url_title = f'{schema.tileMatrixSet} tile matrix set definition' # noqa

                tiling_scheme = LinkType(
                    href=tiling_scheme_url,
                    rel='http://www.opengis.net/def/rel/ogc/1.0/tiling-scheme',
                    type_=tiling_scheme_url_type,
                    title=tiling_scheme_url_title)

        if tiling_scheme is None:
            msg = 'Could not identify a valid tiling schema'
            LOGGER.error(msg)
            raise ProviderConnectionError(msg)

        content = TileSetMetadata(title=title, description=description,
                                  keywords=keywords, crs=crs,
                                  tileMatrixSetURI=tileMatrixSetURI)

        links = []
        service_url_link_type = 'application/vnd.mapbox-vector-tile'
        service_url_link_title = f'{tileset} vector tiles for {layer}'
        service_url_link = LinkType(href=service_url, rel='item',
                                    type_=service_url_link_type,
                                    title=service_url_link_title)

        links.append(tiling_scheme)
        links.append(service_url_link)

        content.links = links

        return content.model_dump(exclude_none=True)

    @staticmethod
    def get_envelope(z, y, x, tileset):
        """
        Calculate the Tile bounding box of a tile at zoom z, y, x.

        WorldCRS84Quad tiles have:
        - origin top-left (y=0 is north)
        - full lon: -180 to 180

        :param tileset: mvt tileset
        :param z: z index
        :param y: y index
        :param x: x index

        :returns: SQL Alchemy Tile Envelope
        """

        if tileset == TileMatrixSetEnum.WORLDCRS84QUAD.value.tileMatrixSet:

            tile_size = 180 / 2 ** z

            xmin = tile_size * x - 180
            ymax = tile_size * -y + 90

            # getting bottom-right coordinates of the tile
            xmax = xmin + tile_size
            ymin = ymax - tile_size

            envelope = ST_MakeEnvelope(xmin, ymin, xmax, ymax, 4326)

        else:
            envelope = ST_TileEnvelope(z, x, y)

        return envelope.label('bounds')

    def __repr__(self):
        return f'<MVTPostgreSQLProvider> {self.data}'
